use crate::identifier::Identifier;
use std::fmt::Debug;
use std::io;
use packet_data::PacketData;
use crate::packet::raw_packet::RawPacket;

/// High level abstraction for packet data.
pub mod packet_data;
/// Low level abstraction for packets.
pub mod raw_packet;

/// Size of the packet header.
pub const PACKET_HEADER_SIZE: usize = 20;
/// Packet size with CRC32.
pub const FULL_PACKET_SIZE: usize = 65559;
/// Packet size without CRC32.
pub const BASE_PACKET_SIZE: usize = 65555;

/// The abstraction for all DomoProto packets.
#[derive(Debug)]
pub struct Packet {
    pub dest: Identifier,
    pub src: Identifier,
    pub packet_id: Identifier,
    pub reply_to: Identifier,
    pub command: u8,
    pub data: PacketData,
}

impl Packet {
    /// Build packet **without** CRC32
    pub fn build_base_packet(&self) -> Vec<u8> {
        match self {
            Packet {
                dest,
                src,
                packet_id,
                reply_to,
                command,
                data,
            } => {
                let mut buf = Vec::new();
                let data = data.clone().into_bytes();
                let data_length = data.len();
                buf.push(0x01);
                buf.extend_from_slice(&dest.bytes);
                buf.extend_from_slice(&src.bytes);
                buf.extend_from_slice(&packet_id.bytes);
                buf.extend_from_slice(&reply_to.bytes);
                buf.push(command.clone());
                buf.push((data_length >> 8) as u8);
                buf.push((data_length & 0xFF) as u8);
                buf.extend(data);
                buf
            }
        }
    }

    pub fn get_crc32(&self) -> u32 {
        crc32fast::hash(self.build_base_packet().as_slice())
    }

    pub fn build_full_packet(&self) -> Vec<u8> {
        match self {
            Packet { .. } => {
                let mut buf = self.build_base_packet();
                buf.extend_from_slice(crc32fast::hash(&buf.as_slice()).to_be_bytes().as_ref());
                buf
            }
        }
    }
}

pub trait ToPacket {
    fn to_packet(self, src: Identifier, dest: Identifier, packet_id: Identifier, reply_to: Identifier) -> Packet;
}

/// Secretly calls RawPacket::try_from().into()
impl TryFrom<Vec<u8>> for Packet {
    type Error = io::Error;
    fn try_from(data: Vec<u8>) -> io::Result<Self> {
        Ok(RawPacket::try_from(data)?.into())
    }
}